#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import requests
import json
from datetime import datetime, timedelta

from octobot_commons.logging.logging_util import get_logger
from octobot_commons.constants import METRICS_URL, COMMUNITY_TOPS_COUNT, METRICS_ROUTE_COMMUNITY
from octobot_commons.config_manager import get_metrics_enabled
from octobot.community.community_fields import CommunityFields

LOGGER = get_logger("CommunityAnalysis")


def get_community_metrics():
    try:
        resp = requests.get(f"{METRICS_URL}{METRICS_ROUTE_COMMUNITY}")
        if resp.status_code != 200:
            LOGGER.error(f"Error when getting community data : error code={resp.status_code}")
        else:
            return _format_community_data(json.loads(resp.text))
    except Exception as e:
        LOGGER.error(f"Error when getting metrics: {e}")


def can_read_metrics(config):
    return get_metrics_enabled(config)


def _format_community_data(json_bot_metrics):
    formatted_community_data = dict()
    formatted_community_data["total_count"] = len(json_bot_metrics)
    formatted_community_data["this_month"] = _get_count_last_months(json_bot_metrics, 1)
    formatted_community_data["last_six_month"] = _get_count_last_months(json_bot_metrics, 6)
    formatted_community_data["top_pairs"] = _get_top_traded_item(json_bot_metrics,
                                                                 CommunityFields.CURRENT_SESSION.value,
                                                                 CommunityFields.PAIRS.value)
    formatted_community_data["top_exchanges"] = _get_top_traded_item(json_bot_metrics,
                                                                     CommunityFields.CURRENT_SESSION.value,
                                                                     CommunityFields.EXCHANGES.value)
    formatted_community_data["top_strategies"] = _get_top_traded_item(json_bot_metrics,
                                                                      CommunityFields.CURRENT_SESSION.value,
                                                                      CommunityFields.EVAL_CONFIG.value)
    return formatted_community_data


def _get_count_last_months(json_bot_metrics, months):
    month_count = 0
    month_min_timestamp = (datetime.now() - timedelta(days=months * 30.5)).timestamp()
    for item in json_bot_metrics:
        if CommunityFields.CURRENT_SESSION.value in item and \
                CommunityFields.UP_TIME.value in item[CommunityFields.CURRENT_SESSION.value] and \
                item[CommunityFields.CURRENT_SESSION.value][CommunityFields.UP_TIME.value] >= month_min_timestamp:
            month_count += 1
    return month_count


def _get_top_traded_item(json_bot_metrics, session_key, key, top_count=COMMUNITY_TOPS_COUNT):
    pair_by_occurrence = _count_occurrences(json_bot_metrics, session_key, key)
    return _get_top_occurrences(pair_by_occurrence, top_count)


def _get_top_occurrences(item_by_occurrence, top_count):
    items = [{"name": key, "count": val} for key, val in item_by_occurrence.items()]
    sorted_items = sorted(items, key=lambda x: x["count"], reverse=True)[:top_count]
    for i, item in enumerate(sorted_items):
        item["rank"] = i + 1
    return sorted_items


def _count_occurrences(items, session_key, key):
    occurrences_by_item = {}
    for item in items:
        if session_key in item and key in item[session_key]:
            for pair in item[session_key][key]:
                if pair in occurrences_by_item:
                    occurrences_by_item[pair] += 1
                else:
                    occurrences_by_item[pair] = 1
    return occurrences_by_item



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot_backtesting.api.backtesting import get_backtesting_data_files

from octobot.api.backtesting import create_independent_backtesting, initialize_and_run_independent_backtesting, \
    join_independent_backtesting, stop_independent_backtesting, log_independent_backtesting_report
from octobot.commands import stop_bot
from octobot.octobot import OctoBot


class OctoBotBacktestingFactory(OctoBot):
    def __init__(self, config, log_report=True, run_on_common_part_only=True):
        super().__init__(config)
        self.independent_backtesting = None
        self.log_report = log_report
        self.run_on_common_part_only = run_on_common_part_only

    async def initialize(self):
        try:
            await self.initializer.create()
            self.independent_backtesting = create_independent_backtesting(
                self.config,
                self.tentacles_setup_config,
                get_backtesting_data_files(self.config),
                run_on_common_part_only=self.run_on_common_part_only)
            await initialize_and_run_independent_backtesting(self.independent_backtesting, log_errors=False)
            await join_independent_backtesting(self.independent_backtesting)
            if self.log_report:
                log_independent_backtesting_report(self.independent_backtesting)
            await stop_independent_backtesting(self.independent_backtesting, memory_check=False)
        except Exception as e:
            self.logger.error(f"Error when starting backtesting: {e}")
        finally:
            self.task_manager.stop_tasks()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time
import uuid

import aiohttp

from octobot.logger import init_octobot_chan_logger
from octobot_services.api.services import stop_services

from octobot.community.community_manager import CommunityManager
from octobot.constants import PROJECT_NAME, LONG_VERSION, CONFIG_KEY, TENTACLES_SETUP_CONFIG_KEY
from octobot.configuration_manager import ConfigurationManager
from octobot.consumers.octobot_channel_consumer import OctoBotChannelGlobalConsumer
from octobot.producers.evaluator_producer import EvaluatorProducer
from octobot.api.octobot_api import OctoBotAPI
from octobot.producers.service_feed_producer import ServiceFeedProducer
from octobot.initializer import Initializer
from octobot.producers.interface_producer import InterfaceProducer
from octobot.producers.exchange_producer import ExchangeProducer
from octobot.task_manager import TaskManager
from octobot_commons.enums import MarkdownFormat
from octobot_commons.logging.logging_util import get_logger
from octobot_services.api.notification import send_notification, create_notification
from octobot_trading.api.exchange import get_exchange_manager_from_exchange_id
from octobot_trading.api.modes import get_trading_modes

"""Main OctoBot class:
- Create all indicators and thread for each cryptocurrencies in config """


class OctoBot:
    """
    Constructor :
    - Load configs
    """

    def __init__(self, config, ignore_config=False, reset_trading_history=False):
        self.start_time = time.time()
        self.config = config
        self.ignore_config = ignore_config
        self.reset_trading_history = reset_trading_history

        # tentacle setup configuration
        self.tentacles_setup_config = None

        # Configuration manager to handle current, edited and startup configurations
        self.configuration_manager = ConfigurationManager()
        self.configuration_manager.add_element(CONFIG_KEY, self.config)

        # Used to know when OctoBot is ready to answer in APIs
        self.initialized = False

        # unique aiohttp session: to be initialized from getter in a task
        self._aiohttp_session = None

        # community if enabled
        self.community_handler = None

        # octobot_api to request the current instance
        self.octobot_api = OctoBotAPI(self)

        # octobot channel global consumer
        self.global_consumer = OctoBotChannelGlobalConsumer(self)

        # octobot instance id
        self.bot_id = str(uuid.uuid4())

        # Logger
        self.logger = get_logger(self.__class__.__name__)

        # Initialize octobot main tools
        self.initializer = Initializer(self)
        self.task_manager = TaskManager(self)

        # Producers
        self.exchange_producer = None
        self.evaluator_producer = None
        self.interface_producer = None
        self.service_feed_producer = None

        self.async_loop = None
        self.community_handler = None

    async def initialize(self):
        await self.initializer.create()
        await self.task_manager.start_tools_tasks()
        await init_octobot_chan_logger(self.bot_id)
        await self.create_producers()
        await self.start_producers()

    async def create_producers(self):
        self.exchange_producer = ExchangeProducer(self.global_consumer.octobot_channel, self,
                                                  None, self.ignore_config)
        self.evaluator_producer = EvaluatorProducer(self.global_consumer.octobot_channel, self)
        self.interface_producer = InterfaceProducer(self.global_consumer.octobot_channel, self)
        self.service_feed_producer = ServiceFeedProducer(self.global_consumer.octobot_channel, self)

    async def start_producers(self):
        await self.evaluator_producer.run()
        await self.exchange_producer.run()
        # Start service feeds now that evaluators registered their feed requirements
        await self.service_feed_producer.run()
        await self.interface_producer.run()
        await self._post_initialize()

    async def _post_initialize(self):
        self.initialized = True

        # make tentacles setup config editable while saving previous states
        self.configuration_manager.add_element(TENTACLES_SETUP_CONFIG_KEY, self.tentacles_setup_config)
        await send_notification(create_notification(f"{PROJECT_NAME} {LONG_VERSION} is starting ...",
                                                    markdown_format=MarkdownFormat.ITALIC))

        # initialize tools
        self._init_community()

    async def stop(self):
        await self.service_feed_producer.stop()
        stop_services()
        await self.interface_producer.stop()
        self.logger.info("Shutting down.")

    def _init_community(self):
        self.community_handler = CommunityManager(self.octobot_api)

    def get_edited_config(self, config_key):
        return self.configuration_manager.get_edited_config(config_key)

    def get_startup_config(self, config_key):
        return self.configuration_manager.get_startup_config(config_key)

    def get_trading_mode(self):
        first_exchange_manager = get_exchange_manager_from_exchange_id(
            next(iter(self.exchange_producer.exchange_manager_ids))
        )
        return get_trading_modes(first_exchange_manager)[0]

    def run_in_main_asyncio_loop(self, coroutine):
        return self.task_manager.run_in_main_asyncio_loop(coroutine)

    def set_watcher(self, watcher):
        self.task_manager.watcher = watcher

    def get_aiohttp_session(self):
        if self._aiohttp_session is None:
            self._aiohttp_session = aiohttp.ClientSession()
        return self._aiohttp_session



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from copy import deepcopy

from octobot.api.backtesting import create_independent_backtesting, \
    initialize_and_run_independent_backtesting, get_independent_backtesting_exchange_manager_ids, \
    join_independent_backtesting
from octobot.strategy_optimizer.test_suite_result import TestSuiteResult
from octobot_backtesting.data import MissingTimeFrame
from octobot.backtesting.abstract_backtesting_test import AbstractBacktestingTest
from octobot_evaluators.constants import CONFIG_FORCED_TIME_FRAME
from octobot_trading.constants import CONFIG_TRADER_RISK, CONFIG_TRADING
from octobot_trading.api.exchange import get_exchange_managers_from_exchange_ids
from octobot_trading.api.profitability import get_profitability_stats
from octobot_trading.api.trades import get_trade_history


class StrategyTestSuite(AbstractBacktestingTest):

    # set to True to skip bigger scenarii and make tests faster
    SKIP_LONG_STEPS = False

    def __init__(self):
        super().__init__()
        self._profitability_results = []
        self._trades_counts = []
        self.current_progress = 0
        self.exceptions = []
        self.evaluators = []

    def get_test_suite_result(self):
        return TestSuiteResult(self._profitability_results,
                               self._trades_counts,
                               self.config[CONFIG_TRADING][CONFIG_TRADER_RISK],
                               self.config[CONFIG_FORCED_TIME_FRAME],
                               self.evaluators,
                               self.strategy_evaluator_class.get_name())

    async def run_test_suite(self, strategy_tester):
        self.exceptions = []
        tests = [self.test_slow_downtrend, self.test_sharp_downtrend, self.test_flat_markets,
                 self.test_slow_uptrend, self.test_sharp_uptrend, self.test_up_then_down]
        print('| ', end='')
        nb_tests = len(tests)
        for i, test in enumerate(tests):
            try:
                await test(strategy_tester)
                self.current_progress = int((i+1)/nb_tests*100)
            except Exception as e:
                self.logger.exception(e, True, f"Exception when running test {test.__name__}: {e}")
                self.exceptions.append(e)
            print('#', end='')
        print(' |', end='')
        return not self.exceptions

    async def test_default_run(self, strategy_tester):
        await strategy_tester.run_test_default_run(None)

    async def test_slow_downtrend(self, strategy_tester):
        await strategy_tester.run_test_slow_downtrend(None, None, None, None, StrategyTestSuite.SKIP_LONG_STEPS)

    async def test_sharp_downtrend(self, strategy_tester):
        await strategy_tester.run_test_sharp_downtrend(None, None, StrategyTestSuite.SKIP_LONG_STEPS)

    async def test_flat_markets(self, strategy_tester):
        await strategy_tester.run_test_flat_markets(None, None, None, None, StrategyTestSuite.SKIP_LONG_STEPS)

    async def test_slow_uptrend(self, strategy_tester):
        await strategy_tester.run_test_slow_uptrend(None, None)

    async def test_sharp_uptrend(self, strategy_tester):
        await strategy_tester.run_test_sharp_uptrend(None, None)

    async def test_up_then_down(self, strategy_tester):
        await strategy_tester.run_test_up_then_down(None, StrategyTestSuite.SKIP_LONG_STEPS)

    def _handle_results(self, independent_backtesting, profitability):
        trades_count = 0
        profitability_result = None
        skip_this_run = False
        if independent_backtesting is not None:
            exchange_manager_ids = get_independent_backtesting_exchange_manager_ids(independent_backtesting)
            try:
                for exchange_manager in get_exchange_managers_from_exchange_ids(exchange_manager_ids):
                    _, profitability, _, market_average_profitability, _ = get_profitability_stats(exchange_manager)
                    # Only one exchange manager per run
                    profitability_result = (profitability, market_average_profitability)
                    trades_count += len(get_trade_history(exchange_manager))
            except (AttributeError, KeyError):
                skip_this_run = True
            if not skip_this_run:
                if profitability_result is None:
                    raise RuntimeError("Error with independent backtesting: no available exchange manager")
                self._profitability_results.append(profitability_result)
                self._trades_counts.append(trades_count)

    async def _run_backtesting_with_current_config(self, data_file_to_use):
        independent_backtesting = None
        try:
            config_to_use = deepcopy(self.config)
            independent_backtesting = create_independent_backtesting(config_to_use,
                                                                     self.tentacles_setup_config,
                                                                     [data_file_to_use],
                                                                     "")
            await initialize_and_run_independent_backtesting(independent_backtesting, log_errors=False)
            await join_independent_backtesting(independent_backtesting)
            return independent_backtesting
        except MissingTimeFrame:
            # ignore this exception: is due to missing of the only required time frame
            return independent_backtesting
        except Exception as e:
            self.logger.exception(e, True, str(e))
            return independent_backtesting



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot.channels.octobot_channel import OctoBotChannelProducer
from octobot.constants import PROJECT_NAME, LONG_VERSION, CONFIG_KEY
from octobot_backtesting.api.backtesting import is_backtesting_enabled
from octobot_services.api.interfaces import create_interface_factory, initialize_global_project_data, is_enabled, \
    start_interfaces, is_enabled_in_backtesting
from octobot_services.api.notification import create_notifier_factory, is_enabled_in_config
from octobot_services.interfaces.util.bot import get_bot_api
from octobot_services.managers.interface_manager import stop_interfaces


class InterfaceProducer(OctoBotChannelProducer):
    """Initializer class:
    - Initialize services, constants and tools
    """

    def __init__(self, channel, octobot):
        super().__init__(channel)
        self.octobot = octobot

        self.interface_list = []
        self.notifier_list = []

    async def start(self):
        in_backtesting = is_backtesting_enabled(self.octobot.config)
        await self._create_interfaces(in_backtesting)
        await self._create_notifiers(in_backtesting)
        await self.start_interfaces()

    async def start_interfaces(self):
        to_start_interfaces = self.interface_list
        started_interfaces = await start_interfaces(to_start_interfaces)
        if len(started_interfaces) != len(to_start_interfaces):
            missing_interfaces = [interface.get_name()
                                  for interface in to_start_interfaces
                                  if interface not in started_interfaces]
            self.logger.error(
                f"{', '.join(missing_interfaces)} interface{'s' if len(missing_interfaces) > 1 else ''} "
                f"did not start properly.")

    async def _create_interfaces(self, in_backtesting):
        # do not overwrite data in case of inner bots init (backtesting)
        if get_bot_api() is None:
            initialize_global_project_data(self.octobot.octobot_api, PROJECT_NAME, LONG_VERSION)
        interface_factory = create_interface_factory(self.octobot.config)
        interface_list = interface_factory.get_available_interfaces()
        for interface_class in interface_list:
            await self._create_interface_if_relevant(interface_factory, interface_class, in_backtesting,
                                                     self.octobot.get_edited_config(CONFIG_KEY))

    async def _create_notifiers(self, in_backtesting):
        notifier_factory = create_notifier_factory(self.octobot.config)
        notifier_list = notifier_factory.get_available_notifiers()
        for notifier_class in notifier_list:
            await self._create_notifier_class_if_relevant(notifier_factory, notifier_class, in_backtesting,
                                                          self.octobot.get_edited_config(CONFIG_KEY))

    async def _create_interface_if_relevant(self, interface_factory, interface_class,
                                            backtesting_enabled, edited_config):
        if self._is_interface_relevant(interface_class, backtesting_enabled):
            interface_instance = await interface_factory.create_interface(interface_class)
            await interface_instance.initialize(backtesting_enabled, edited_config)
            self.interface_list.append(interface_instance)

    async def _create_notifier_class_if_relevant(self, notifier_factory, notifier_class,
                                                 backtesting_enabled, edited_config):
        if self._is_notifier_relevant(notifier_class, backtesting_enabled):
            notifier_instance = await notifier_factory.create_notifier(notifier_class)
            await notifier_instance.initialize(backtesting_enabled, edited_config)
            self.notifier_list.append(notifier_instance)

    def _is_interface_relevant(self, interface_class, backtesting_enabled):
        return is_enabled(interface_class) and \
               all(service.get_is_enabled(self.octobot.config) for service in interface_class.REQUIRED_SERVICES) and \
               (not backtesting_enabled or (backtesting_enabled and is_enabled_in_backtesting(interface_class)))

    def _is_notifier_relevant(self, notifier_class, backtesting_enabled):
        return is_enabled_in_config(notifier_class, self.octobot.config) and \
               all(service.get_is_enabled(self.octobot.config)
                   for service in notifier_class.REQUIRED_SERVICES) and \
               not backtesting_enabled

    async def stop(self):
        await stop_interfaces(self.interface_list)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

PROJECT_NAME = "OctoBot"
SHORT_VERSION = "0.4.0"  # major.minor.revision
PATCH_VERSION = ""  # patch : pX
VERSION_DEV_PHASE = "a"  # alpha : a / beta : b / release candidate : rc
VERSION_PHASE = "12"  # XX
VERSION = f"{SHORT_VERSION}{VERSION_DEV_PHASE}{VERSION_PHASE}"
LONG_VERSION = f"{SHORT_VERSION}{PATCH_VERSION}{VERSION_DEV_PHASE}{VERSION_PHASE}"

# OctoBot urls
OCTOBOT_ONLINE = "https://www.tentacles.octobot.online"
REPOSITORY = "repository"
TENTACLES_REPOSITORY = "tentacles"
OFFICIALS = "officials"
TENTACLE_BASE_CATEGORY = "base"

# tentacles
ENV_TENTACLES_URL = "TENTACLES_URL"
ENV_TENTACLES_URL_TAG = "TENTACLES_URL_TAG"
ENV_TENTACLES_URL_CATEGORY = "TENTACLES_URL_CATEGORY"
ENV_TENTACLES_URL_SUBCATEGORY = "TENTACLES_URL_SUBCATEGORY"
DEFAULT_TENTACLES_URL = os.getenv(ENV_TENTACLES_URL,
                                  f"{OCTOBOT_ONLINE}/{REPOSITORY}/{TENTACLES_REPOSITORY}/{OFFICIALS}/"
                                  f"{os.getenv(ENV_TENTACLES_URL_CATEGORY, TENTACLE_BASE_CATEGORY)}/"
                                  f"{os.getenv(ENV_TENTACLES_URL_SUBCATEGORY, '')}"
                                  f"{os.getenv(ENV_TENTACLES_URL_TAG, LONG_VERSION)}.zip")
DEFAULT_TENTACLES_PACKAGE_NAME = "OctoBot-Default-Tentacles"

# logs
LOGS_FOLDER = "logs"

# config types keys
CONFIG_KEY = "config"
TENTACLES_SETUP_CONFIG_KEY = "tentacles_setup"

# terms of service
CONFIG_ACCEPTED_TERMS = "accepted_terms"

# DEBUG
CONFIG_DEBUG_OPTION = "DEV-MODE"
FORCE_ASYNCIO_DEBUG_OPTION = False

# Files
CONFIG_FILE = "config.json"
TEMP_RESTORE_CONFIG_FILE = "temp_config.json"
CONFIG_FOLDER = "config"
SCHEMA = "schema"
CONFIG_FILE_SCHEMA = f"{CONFIG_FOLDER}/config_{SCHEMA}.json"
DEFAULT_CONFIG_FILE = f"{CONFIG_FOLDER}/default_config.json"
LOGGING_CONFIG_FILE = f"{CONFIG_FOLDER}/logging_config.ini"
LOG_FILE = f"{LOGS_FOLDER}/{PROJECT_NAME}.log"

# Optimizer
OPTIMIZER_FORCE_ASYNCIO_DEBUG_OPTION = False

# Channel
OCTOBOT_CHANNEL = "OctoBot"

OCTOBOT_KEY = b'uVEw_JJe7uiXepaU_DR4T-ThkjZlDn8Pzl8hYPIv7w0='



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot.constants import CONFIG_KEY, TENTACLES_SETUP_CONFIG_KEY
from octobot.commands import stop_bot as command_stop, restart_bot as command_restart


class OctoBotAPI:

    def __init__(self, octobot):
        self._octobot = octobot

    def is_initialized(self) -> bool:
        return self._octobot.initialized

    def get_exchange_manager_ids(self) -> list:
        return self._octobot.exchange_producer.exchange_manager_ids

    def get_global_config(self) -> dict:
        return self._octobot.config

    def get_startup_config(self) -> object:
        return self._octobot.get_startup_config(CONFIG_KEY)

    def get_edited_config(self) -> object:
        return self._octobot.get_edited_config(CONFIG_KEY)

    def get_startup_tentacles_config(self) -> object:
        return self._octobot.get_startup_config(TENTACLES_SETUP_CONFIG_KEY)

    def get_edited_tentacles_config(self) -> object:
        return self._octobot.get_edited_config(TENTACLES_SETUP_CONFIG_KEY)

    def get_trading_mode(self) -> object:
        return self._octobot.get_trading_mode()

    def get_tentacles_setup_config(self) -> object:
        return self._octobot.tentacles_setup_config

    def get_start_time(self) -> float:
        return self._octobot.start_time

    def get_matrix_id(self) -> str:
        return self._octobot.evaluator_producer.matrix_id

    def get_aiohttp_session(self) -> object:
        return self._octobot.get_aiohttp_session()

    def run_in_main_asyncio_loop(self, coroutine):
        return self._octobot.run_in_main_asyncio_loop(coroutine)

    def stop_tasks(self) -> None:
        self._octobot.task_manager.stop_tasks()

    def stop_bot(self) -> None:
        command_stop(self._octobot)

    @staticmethod
    def restart_bot() -> None:
        command_restart()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import uuid
import gc
from sys import getrefcount
from asyncio import get_event_loop

from octobot_backtesting.api.backtesting import stop_backtesting, initialize_backtesting, start_backtesting, \
    get_importers, adapt_backtesting_channels
from octobot_backtesting.api.importer import stop_importer
from octobot_backtesting.importers.exchanges.exchange_importer import ExchangeDataImporter
from octobot_commons.logging.logging_util import get_logger
from octobot_trading.api.exchange import get_exchange_managers_from_exchange_ids, stop_exchange
from octobot_evaluators.api.matrix import del_matrix
from octobot_evaluators.api.evaluators import stop_evaluator, stop_all_evaluator_channels, initialize_evaluators
from octobot_evaluators.api.initialization import del_evaluator_channels, create_evaluator_channels
from octobot_services.api.service_feeds import stop_service_feed, start_service_feed
from octobot_trading.api.modes import init_trading_mode_config
from octobot_trading.exchanges.data.exchange_symbol_data import ExchangeSymbolData
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.exchanges.exchange_simulator import ExchangeSimulator
from octobot_trading.producers.simulator import OHLCVUpdaterSimulator
from octobot_evaluators.api.evaluators import create_all_type_evaluators
from octobot_trading.api.exchange import get_exchange_configuration_from_exchange_id, create_exchange_builder, \
    get_exchange_manager_id
from octobot.logger import init_exchange_chan_logger, init_evaluator_chan_logger, BOT_CHANNEL_LOGGER


class OctoBotBacktesting:

    def __init__(self, backtesting_config,
                 tentacles_setup_config,
                 symbols_to_create_exchange_classes,
                 backtesting_files,
                 run_on_common_part_only):
        self.logger = get_logger(self.__class__.__name__)
        self.backtesting_config = backtesting_config
        self.tentacles_setup_config = tentacles_setup_config
        self.bot_id = str(uuid.uuid4())
        self.matrix_id = ""
        self.exchange_manager_ids = []
        self.symbols_to_create_exchange_classes = symbols_to_create_exchange_classes
        self.evaluators = []
        self.service_feeds = []
        self.backtesting_files = backtesting_files
        self.backtesting = None
        self.run_on_common_part_only = run_on_common_part_only

    async def initialize_and_run(self):
        self.logger.info(f"Starting on {self.backtesting_files} with {self.symbols_to_create_exchange_classes}")
        await self._init_evaluators()
        await self._init_service_feeds()
        await self._init_exchanges()
        await self._create_evaluators()
        await self._create_service_feeds()
        await start_backtesting(self.backtesting)
        if BOT_CHANNEL_LOGGER is not None:
            await self.start_loggers()

    async def stop(self, memory_check=False):
        self.logger.info(f"Stopping for {self.backtesting_files} with {self.symbols_to_create_exchange_classes}")
        try:
            await stop_backtesting(self.backtesting)
            exchange_managers = []
            try:
                for exchange_manager in get_exchange_managers_from_exchange_ids(self.exchange_manager_ids):
                    exchange_managers.append(exchange_manager)
                    await stop_exchange(exchange_manager)
            except KeyError:
                # exchange managers are not added in global exchange list when an exception occurred
                pass
            for evaluators in self.evaluators:
                # evaluators by type
                for evaluator in evaluators:
                    # evaluator instance
                    if evaluator is not None:
                        await stop_evaluator(evaluator)
            await stop_all_evaluator_channels(self.matrix_id)
            del_evaluator_channels(self.matrix_id)
            del_matrix(self.matrix_id)
            for service_feed in self.service_feeds:
                await stop_service_feed(service_feed)

            for importer in get_importers(self.backtesting):
                await stop_importer(importer)

            if memory_check:
                to_reference_check = exchange_managers + [self.backtesting]
                # Call at the next loop iteration to first let coroutines get cancelled
                # (references to coroutine and caller objects are kept while in async loop)
                get_event_loop().call_soon(self.memory_leak_checkup, to_reference_check)
            self.backtesting = None
        except Exception as e:
            self.logger.exception(e, True, f"Error when stopping independent backtesting: {e}")

    def memory_leak_checkup(self, to_check_elements):
        self.logger.debug(f"Memory leak checking {[e.__class__.__name__ for e in to_check_elements]}")
        memory_leak_errors = []
        for i in range(len(to_check_elements)):
            if getrefcount(to_check_elements[i]) > 2:
                # Using PyCharm debugger, right click on the element variable and use "Find references"
                # Warning: Python debugger can add references when watching an element
                element = to_check_elements[i]
                # Now expect 3 references because the above element variable adds a reference
                memory_leak_errors.append(f" Too many remaining references on the {element.__class__.__name__} element "
                                          f"after {self.__class__.__name__} run, the garbage collector won't free it "
                                          f"(expected a maximum of 3 references): {getrefcount(element)} actual "
                                          f"references ({element})")
        if memory_leak_errors:
            errors = "\n".join(memory_leak_errors)
            raise AssertionError(
                f"[Dev oriented error: no effect on backtesting result, please report if you see it]: {errors}"
            )

    # Use check_remaining_objects to check remaining objects from garbage collector after calling stop().
    # Warning: can take a long time when a lot of objects exist
    def check_remaining_objects(self):
        objects_leak_errors = []
        exchanges_count = len(self.exchange_manager_ids)
        to_watch_objects = (ExchangeSymbolData, ExchangeManager, ExchangeSimulator, OHLCVUpdaterSimulator)
        objects_references = {obj: (0, []) for obj in to_watch_objects}
        expected_max_objects_references = {
            ExchangeSymbolData: exchanges_count + 1,
            ExchangeManager: exchanges_count + 1,
            ExchangeSimulator: exchanges_count,
            OHLCVUpdaterSimulator: exchanges_count
        }
        for obj in gc.get_objects():
            if isinstance(obj, to_watch_objects):
                objects_references[type(obj)][1].append(obj)
                objects_references[type(obj)] = (objects_references[type(obj)][0] + 1,
                                                 objects_references[type(obj)][1])

        for obj, max_ref in expected_max_objects_references.items():
            if objects_references[obj][0] > max_ref:
                objects_leak_errors.append(_get_remaining_object_error(obj,
                                                                       max_ref,
                                                                       objects_references[obj]))

        if objects_leak_errors:
            errors = "\n".join(objects_leak_errors)
            raise AssertionError(
                f"[Dev oriented error: no effect on backtesting result, please report if you see it]: {errors}"
            )

    async def _init_evaluators(self):
        self.matrix_id = await initialize_evaluators(self.backtesting_config, self.tentacles_setup_config)
        await create_evaluator_channels(self.matrix_id, is_backtesting=True)

    async def _init_service_feeds(self):
        try:
            from octobot_services.api.service_feeds import create_service_feed_factory
            service_feed_factory = create_service_feed_factory(self.backtesting_config,
                                                               get_event_loop(),
                                                               self.bot_id)
            self.service_feeds = [service_feed_factory.create_service_feed(feed)
                                  for feed in service_feed_factory.get_available_service_feeds(True)]
        except ImportError:
            self.logger.error("Octobot Services package is required to test service feeds. Service feeds will not "
                              "be available for this backtesting.")

    async def _create_evaluators(self):
        for exchange_id in self.exchange_manager_ids:
            exchange_configuration = get_exchange_configuration_from_exchange_id(exchange_id)
            self.evaluators = await create_all_type_evaluators(
                self.backtesting_config,
                self.tentacles_setup_config,
                matrix_id=self.matrix_id,
                exchange_name=exchange_configuration.exchange_name,
                bot_id=self.bot_id,
                symbols_by_crypto_currencies=exchange_configuration.symbols_by_crypto_currencies,
                symbols=exchange_configuration.symbols,
                time_frames=exchange_configuration.time_frames_without_real_time,
                real_time_time_frames=exchange_configuration.real_time_time_frames)

    async def _create_service_feeds(self):
        for feed in self.service_feeds:
            if not await start_service_feed(feed, False, {}):
                self.logger.error(f"Failed to start {feed.get_name()}. Evaluators requiring this service feed "
                                  f"might not work properly")

    async def _init_exchanges(self):
        init_trading_mode_config(self.backtesting_config)
        self.backtesting = await initialize_backtesting(self.backtesting_config,
                                                        exchange_ids=self.exchange_manager_ids,
                                                        matrix_id=self.matrix_id,
                                                        data_files=self.backtesting_files)
        # modify_backtesting_channels before creating exchanges as they require the current backtesting time to
        # initialize
        await adapt_backtesting_channels(self.backtesting,
                                         self.backtesting_config,
                                         ExchangeDataImporter,
                                         run_on_common_part_only=self.run_on_common_part_only)

        for exchange_class_string in self.symbols_to_create_exchange_classes.keys():
            exchange_builder = create_exchange_builder(self.backtesting_config, exchange_class_string) \
                .has_matrix(self.matrix_id) \
                .use_tentacles_setup_config(self.tentacles_setup_config) \
                .set_bot_id(self.bot_id) \
                .is_simulated() \
                .is_rest_only() \
                .is_backtesting(self.backtesting)
            try:
                await exchange_builder.build()
            finally:
                # always save exchange manager ids and backtesting instances
                self.exchange_manager_ids.append(get_exchange_manager_id(exchange_builder.exchange_manager))

    async def start_loggers(self):
        await self.start_exchange_loggers()
        await init_evaluator_chan_logger(self.matrix_id)

    async def start_exchange_loggers(self):
        for exchange_manager_id in self.exchange_manager_ids:
            await init_exchange_chan_logger(exchange_manager_id)


def _get_remaining_object_error(obj, expected, actual):
    error = f"too many remaining {obj.__name__} instances: expected: {expected} actual {actual[0]}"
    for i in range(len(actual[1])):
        error += f"{getrefcount(actual[1][i])} references on {actual[1][i]}"
        return error



# cython: language_level=3
#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cdef class OctoBotAPI:

    cdef object _octobot

    cpdef bint is_initialized(self)
    cpdef list get_exchange_manager_ids(self)
    cpdef dict get_global_config(self)
    cpdef object get_startup_tentacles_config(self)
    cpdef object get_edited_tentacles_config(self)
    cpdef object get_startup_config(self)
    cpdef object get_edited_config(self)
    cpdef object get_trading_mode(self)
    cpdef double get_start_time(self)
    cpdef str get_matrix_id(self)
    cpdef object get_aiohttp_session(self)
    cpdef object run_in_main_asyncio_loop(self, object coroutine)
    cpdef void stop_tasks(self)
    cpdef void stop_bot(self)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
from abc import ABCMeta, ABC
import copy

from octobot.backtesting.abstract_backtesting_test import AbstractBacktestingTest
from octobot.api.backtesting import get_independent_backtesting_exchange_manager_ids, \
    create_independent_backtesting, initialize_and_run_independent_backtesting, join_independent_backtesting
from octobot_commons.constants import CONFIG_WILDCARD
from octobot_commons.logging.logging_util import get_logger
from octobot_commons.tentacles_management.class_inspector import get_class_from_string, evaluator_parent_inspection, \
    trading_mode_parent_inspection
from octobot_commons.tests.test_config import load_test_config
from octobot_evaluators.evaluator import StrategyEvaluator
from octobot_tentacles_manager.api.configurator import update_activation_configuration, get_tentacles_activation
from octobot_trading.api.exchange import get_exchange_managers_from_exchange_ids
from octobot_trading.api.profitability import get_profitability_stats
from octobot_trading.constants import CONFIG_SIMULATOR_FEES_TAKER, CONFIG_SIMULATOR_FEES_MAKER, CONFIG_SIMULATOR_FEES, \
    CONFIG_SIMULATOR
from octobot_trading.modes import AbstractTradingMode
from tests.test_utils.config import load_test_tentacles_config
from tentacles.Trading import Mode
from tentacles.Evaluator import Strategies
from tentacles.Trading.Mode import DailyTradingMode


DEFAULT_SYMBOL = "ICX/BTC"
DATA_FILE_PATH = "tests/static/"


class AbstractStrategyTest(AbstractBacktestingTest, ABC):
    __metaclass__ = ABCMeta

    def initialize(self, strategy_evaluator_class, trading_mode_class=DailyTradingMode, config=None):
        self.logger = get_logger(self.__class__.__name__)
        self.config = load_test_config() if config is None else config
        # remove fees
        self.config[CONFIG_SIMULATOR][CONFIG_SIMULATOR_FEES][CONFIG_SIMULATOR_FEES_TAKER] = 0
        self.config[CONFIG_SIMULATOR][CONFIG_SIMULATOR_FEES][CONFIG_SIMULATOR_FEES_MAKER] = 0
        self.strategy_evaluator_class = strategy_evaluator_class
        self.trading_mode_class = trading_mode_class
        self.tentacles_setup_config = load_test_tentacles_config()
        self._update_tentacles_config(strategy_evaluator_class, trading_mode_class)

    def _handle_results(self, independent_backtesting, profitability):
        exchange_manager_ids = get_independent_backtesting_exchange_manager_ids(independent_backtesting)
        for exchange_manager in get_exchange_managers_from_exchange_ids(exchange_manager_ids):
            _, run_profitability, _, market_average_profitability, _ = get_profitability_stats(exchange_manager)
            actual = round(run_profitability, 3)
            # uncomment this print for building tests
            # print(f"results: rounded run profitability {actual} market profitability: {market_average_profitability}"
            #       f" expected: {profitability} [result: {actual ==  profitability}]")
            assert actual == profitability

    async def _run_backtesting_with_current_config(self, data_file_to_use):
        config_to_use = copy.deepcopy(self.config)
        independent_backtesting = create_independent_backtesting(config_to_use,
                                                                 self.tentacles_setup_config,
                                                                 [data_file_to_use],
                                                                 "")
        await initialize_and_run_independent_backtesting(independent_backtesting, log_errors=False)
        await join_independent_backtesting(independent_backtesting)
        return independent_backtesting

    def _update_tentacles_config(self, strategy_evaluator_class, trading_mode_class):
        default_evaluators = strategy_evaluator_class.get_default_evaluators()
        to_update_config = {}
        for tentacle_class_name in get_tentacles_activation(self.tentacles_setup_config):
            if CONFIG_WILDCARD not in default_evaluators and tentacle_class_name in default_evaluators:
                to_update_config[tentacle_class_name] = True
            elif get_class_from_string(tentacle_class_name, StrategyEvaluator, Strategies,
                                       evaluator_parent_inspection) is not None or \
                get_class_from_string(tentacle_class_name, AbstractTradingMode, Mode,
                                      trading_mode_parent_inspection) is not None:
                to_update_config[tentacle_class_name] = False
            elif CONFIG_WILDCARD not in default_evaluators:
                to_update_config[tentacle_class_name] = False
        # Add required elements if missing
        to_update_config.update({evaluator: True for evaluator in default_evaluators})
        to_update_config[strategy_evaluator_class.get_name()] = True
        to_update_config[trading_mode_class.get_name()] = True
        update_activation_configuration(self.tentacles_setup_config, to_update_config, False, add_missing_elements=True)



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from octobot.channels.octobot_channel import OctoBotChannelProducer
from octobot.logger import init_evaluator_chan_logger
from octobot_backtesting.api.backtesting import is_backtesting_enabled
from octobot_commons.enums import OctoBotChannelSubjects
from octobot_evaluators.api.evaluators import initialize_evaluators
from octobot_evaluators.api.initialization import create_evaluator_channels
from octobot_evaluators.consumers.octobot_channel_consumer import OctoBotChannelEvaluatorActions as EvaluatorActions, \
    OctoBotChannelEvaluatorDataKeys as EvaluatorKeys


class EvaluatorProducer(OctoBotChannelProducer):
    """EvaluatorFactory class:
    - Create evaluators
    """

    def __init__(self, channel, octobot):
        super().__init__(channel)
        self.octobot = octobot
        self.tentacles_setup_config = self.octobot.tentacles_setup_config

        self.matrix_id = None

    async def start(self):
        self.matrix_id = await initialize_evaluators(self.octobot.config, self.tentacles_setup_config)
        await create_evaluator_channels(self.matrix_id, is_backtesting=is_backtesting_enabled(self.octobot.config))
        await init_evaluator_chan_logger(self.matrix_id)

    async def create_evaluators(self, exchange_configuration):
        await self.send(bot_id=self.octobot.bot_id,
                        subject=OctoBotChannelSubjects.CREATION.value,
                        action=EvaluatorActions.EVALUATOR.value,
                        data={
                            EvaluatorKeys.TENTACLES_SETUP_CONFIG.value: self.octobot.tentacles_setup_config,
                            EvaluatorKeys.MATRIX_ID.value: self.octobot.evaluator_producer.matrix_id,
                            EvaluatorKeys.EXCHANGE_CONFIGURATION.value: exchange_configuration,
                            EvaluatorKeys.CONFIG.value: self.octobot.config
                        })





#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import copy

from octobot_commons.config import load_config, get_user_config
from octobot_commons.config_manager import _handle_encrypted_value, dump_json, save_config
from octobot_commons.constants import TEMP_RESTORE_CONFIG_FILE, CONFIG_ENABLED_OPTION
from octobot_commons.logging.logging_util import get_logger
from octobot_trading.api.trader import is_trader_enabled_in_config, is_trader_simulator_enabled_in_config
from octobot_trading.constants import CONFIG_EXCHANGES, CONFIG_EXCHANGE_ENCRYPTED_VALUES, CONFIG_SIMULATOR, \
    CONFIG_TRADER


class ConfigurationManager:
    def __init__(self):
        self.configuration_elements = {}

    def add_element(self, key, element):
        self.configuration_elements[key] = ConfigurationElement(element)

    def get_edited_config(self, key):
        return self.configuration_elements[key].edited_config

    def get_startup_config(self, key):
        return self.configuration_elements[key].startup_config


class ConfigurationElement:
    def __init__(self, element):
        self.config = element
        self.startup_config = copy.deepcopy(element)
        self.edited_config = copy.deepcopy(element)


def config_health_check(config):
    # 1 ensure api key encryption
    should_replace_config = False
    if CONFIG_EXCHANGES in config:
        for exchange, exchange_config in config[CONFIG_EXCHANGES].items():
            for key in CONFIG_EXCHANGE_ENCRYPTED_VALUES:
                try:
                    if not _handle_encrypted_value(key, exchange_config, verbose=True):
                        should_replace_config = True
                except Exception as e:
                    get_logger().exception(e, True, f"Exception when checking exchange config encryption: {e}")

    # 2 ensure single trader activated
    try:
        trader_enabled = is_trader_enabled_in_config(config)
        if trader_enabled:
            simulator_enabled = is_trader_simulator_enabled_in_config(config)
            if simulator_enabled:
                get_logger().error(f"Impossible to activate a trader simulator additionally to a real trader, "
                                   f"simulator deactivated.")
                config[CONFIG_SIMULATOR][CONFIG_ENABLED_OPTION] = False
                should_replace_config = True
    except KeyError as e:
        get_logger().exception(e, True, f"KeyError when checking traders activation: {e}. Activating trader simulator.")
        config[CONFIG_SIMULATOR][CONFIG_ENABLED_OPTION] = True
        config[CONFIG_TRADER][CONFIG_ENABLED_OPTION] = False
        should_replace_config = True

    # 3 save fixed config if necessary
    if should_replace_config:
        try:
            save_config(get_user_config(), config, TEMP_RESTORE_CONFIG_FILE, json_data=dump_json(config))
            return config
        except Exception as e:
            get_logger().error(f"Save of the health checked config failed : {e}, will use the initial config")
            return load_config(error=False, fill_missing_fields=True)
